{
  "type": "record",
  "name": "TagKey",
  "namespace": "com.linkedin.pegasus2avro.metadata.key",
  "doc": "Key for a Tag",
  "fields": [
    {
      "name": "name",
      "type": "string",
      "doc": "The tag name, which serves as a unique id",
      "Searchable": {
        "boostScore": 10.0,
        "enableAutocomplete": true,
        "fieldName": "id",
        "fieldType": "TEXT_PARTIAL"
      }
    }
  ],
  "Aspect": {
    "name": "tagKey",
    "keyForEntity": "tag",
    "entityCategory": "_unset_",
    "entityAspects": [
      "tagProperties",
      "ownership",
      "deprecation"
    ]
  }
}